
#include "TSystem.h"
#include "TCanvas.h"
#include "TSeqCollection.h"
#include "Buttons.h"
#include "TGraph2DErrors.h"
#include "TVirtualFitter.h"
#include "TPolyLine3D.h"
#include "TMath.h"
#include "TVector3.h"

#include <iostream>
#include <cstdlib>
#include <vector>

#include "CAnalysis.hh"

ClassImp(CAnalysis)

CAnalysis::CAnalysis(){/* nothing */;}

CAnalysis::~CAnalysis(){/* nothing */;}

//----------------------------------------------------------------------------//

FitPointType::size_type CAnalysis::FitPointNumber;

void CAnalysis::LinearFit(const FitPointType& x, const FitPointType& y,
			  const FitPointType& ysigma, double& a, double& b)
{
  FitPointNumber = x.size();
  if(FitPointNumber != y.size() || FitPointNumber != ysigma.size()){
    std::cout << "size of x,y,ysigma must be same" << std::endl;
    std::abort();
  }

  FitPointType xx = MultiplyVector(x, x);
  FitPointType xy = MultiplyVector(x, y);

  double sbu = SB(ysigma);
  double sbx = SB(x, ysigma);
  double sby = SB(y, ysigma);
  double sbxx = SB(xx, ysigma);
  double sbxy = SB(xy, ysigma);
  double wmx = sbx/sbu;
  double wmy = sby/sbu;

  b = (sbu*sbxy-sbx*sby)/(sbu*sbxx-sbx*sbx);
  a = wmy-b*wmx;
}

FitPointType CAnalysis::MultiplyVector(FitPointType x, FitPointType y)
{
  FitPointType xy;

  for(FitPointType::size_type i = 0; i < FitPointNumber; ++i){
    xy.push_back(x[i]*y[i]);
  }

  return xy;
}

double CAnalysis::SD(const FitPointType& x, const FitPointType& y,
		     double a, double b)
{
  double sum(0);
  for(FitPointType::size_type i = 0; i < FitPointNumber; ++i){
    double residual = y[i]-(a+b*x[i]);
    sum += residual*residual;
  }

  return sum/double(FitPointNumber);
}

double CAnalysis::SB(const FitPointType& f, const FitPointType& ysigma)
{
  double sum(0);
  for(FitPointType::size_type i = 0; i < FitPointNumber; ++i){
    sum += f[i]/ysigma[i]/ysigma[i];
  }  

  return sum/double(FitPointNumber);
}

double CAnalysis::SB(const FitPointType& ysigma)
{
  double sum(0);
  for(FitPointType::size_type i = 0; i < FitPointNumber; ++i){
    sum += 1./ysigma[i]/ysigma[i];
  }  

  return sum/double(FitPointNumber);
}

//----------------------------------------------------------------------------//

void CAnalysis::line(double t, double *p, double &x, double &y, double &z)
{ 
  x = p[0] + p[1]*t; 
  y = p[2] + p[3]*t;
  z = t; 
} 

double CAnalysis::distance2_3d(double x,double y,double z, double ex, double ey, double ez, double *p)
{ 
  TVector3 xp(x,y,z); 
  TVector3 x0(p[0], p[2], 0.); 
  TVector3 v0(p[1], p[3], 1.); 
  Double_t k = v0.Dot(xp-x0)/v0.Mag2();

  TVector3 v1 = (x0-xp)+k*v0;//x,y,zと直線を結ぶ、直線に対して垂直なベクトル

  TVector3 v2(v1.x()/ex, v1.y()/ey, v1.z()/ez);
  return v2.Mag2();
}

double CAnalysis::distance2_2d(double x,double y,double z, double ex, double ey, double /*ez*/, double *p)
{ 
  double t = z;
  double xt, yt, zt;
  line(t, p, xt, yt, zt);
  double d2 = pow((xt-x)/ex, 2.) + pow((yt-y)/ey, 2.);
  return d2; 
}

double CAnalysis::distance2(double x,double y,double z, double /*ex*/, double /*ey*/, double /*ez*/, double *p)
{ 
  TVector3 xp(x,y,z); 
  TVector3 x0(p[0], p[2], 0. ); 
  TVector3 x1(p[0] + p[1], p[2] + p[3], 1. ); 
  TVector3 u = (x1-x0).Unit(); 
  double d2 = ((xp-x0).Cross(u)).Mag2(); 
  return d2; 
}

void CAnalysis::SumDistance2(int &, double *, double & sum, double * par, int )
{ 
  TGraph2DErrors* gr = dynamic_cast<TGraph2DErrors*>((TVirtualFitter::GetFitter())->GetObjectFit());
  //  assert(gr != 0);
  double* x = gr->GetX();
  double* y = gr->GetY();
  double* z = gr->GetZ();
  double* ex = gr->GetEX();
  double* ey = gr->GetEY();
  double* ez = gr->GetEZ();
  int npoints = gr->GetN();
  sum = 0;
  for (int i  = 0; i < npoints; ++i) { 
    double d = distance2_3d(x[i],y[i],z[i], ex[i], ey[i], ez[i], par); 
    sum += d;
  }
}

double CAnalysis::line3Dfit(TGraph2DErrors* gr, double parFit[4])
{
  TVirtualFitter* min = TVirtualFitter::Fitter(0,4);
  min->SetObjectFit(gr);
  min->SetFCN(SumDistance2);
   
  Double_t arglist[10];
  arglist[0] = -1;
  min->ExecuteCommand("SET PRINT",arglist,1);
  
  double pStart[4] = {1,1,1,1};
  min->SetParameter(0,"x0",pStart[0],0.01,0,0);
  min->SetParameter(1,"Ax",pStart[1],0.01,0,0);
  min->SetParameter(2,"y0",pStart[2],0.01,0,0);
  min->SetParameter(3,"Ay",pStart[3],0.01,0,0);
    
  arglist[0] = 1000; // number of function calls 
  arglist[1] = 0.001; // tolerance 
  min->ExecuteCommand("MIGRAD",arglist,2);

  //if (minos) min->ExecuteCommand("MINOS",arglist,0);
  int nvpar,nparx; 
  double amin,edm, errdef;
  min->GetStats(amin,edm,errdef,nvpar,nparx);

  //  min->PrintResults(1,amin);
  //  gr->Draw("p0");

  // get fit parameters
  for (int i = 0; i <4; ++i) parFit[i] = min->GetParameter(i);

  return amin;
}

void CAnalysis::DrawLine(double parFit[4], Color_t color, double t0, double dt)
{
  int n = 10;
  TPolyLine3D* l = new TPolyLine3D(n);

  for(int i = 0; i<n+1; ++i){
    double t = t0+ dt*i/n;
    double xtemp, ytemp, ztemp;
    line(t,parFit,xtemp,ytemp,ztemp);
    l->SetPoint(i,xtemp,ytemp,ztemp);
  }

  l->SetLineColor(color);
  l->Draw("same");
}

void CAnalysis::EventLoop()
{
  TSeqCollection* tsc = gROOT->GetListOfCanvases();
  TCanvas* canvas = (TCanvas*)tsc->Last();

  canvas->HandleInput(kNoEvent, 0, 0);
  while (!gSystem->ProcessEvents() && canvas) {
    canvas = (TCanvas*)gROOT->GetSelectedPad();
    Int_t event = canvas->GetEvent();
    if(event == kKeyPress) break;
    if(gSystem->ProcessEvents()) break;
    gSystem->Sleep(100);
  }
}

