#include <iostream>
#include <iomanip>
#include <cmath>

#include "TROOT.h"
#include "TRandom3.h"
#include "TMath.h"

#include "CKinema.hh"
#include "CParticle.hh"

ClassImp(CKinema)

const double CKinema::fAtomicMass(931.494);
const double CKinema::fLightVelocity(29.979);

TRandom3 CKinema::random; 

CKinema::CKinema()
{
  /* nothing */;
}

CKinema::~CKinema()
{
  /* nothing */;
}

void CKinema::LorentzBoost(CParticle& particle, double beta, int i)
{
  double p[3];
  for(int j = 0 ; j < 3; ++j){
    p[j] = particle.p(int(j));
  }

  double gamma = sqrt(1/(1+beta)/(1-beta));
  p[i] = -gamma*beta*particle.e() + gamma*particle.p(i);

  particle.p(p[0], p[1], p[2]);
}

void CKinema::Fission
(CParticle& particle1, CParticle& particle2, double erel, double ltheta, double lphi)
{
  double M = erel + InvariantMass(particle1, particle2);

  double e1 = (M*M - particle2.m()*particle2.m() + particle1.m()*particle1.m())/(2*M);
  double e2 = (M*M - particle1.m()*particle1.m() + particle2.m()*particle2.m())/(2*M);

  particle1.e(e1);
  particle1.thetaphi(ltheta, lphi);

  particle2.e(e2);
  particle2.thetaphibar(ltheta, lphi);
}

void CKinema::Fission(CParticle& particle1, CParticle& particle2, double etotal)
{
  double pabs = sqrt(pow((particle1.m2() - particle2.m2()), 2)
		   - 2.*etotal*etotal*(particle1.m2() + particle2.m2())
		   +pow(etotal, 4))/(2.*etotal);

  particle1.pabs(pabs);
  particle2.pabs(pabs);
}

void CKinema::FissionErel(CParticle& particle1, CParticle& particle2, double erel, double kin)
{
  CParticle pv(fAtomicMass);
  pv.k(kin);
  double beta = pv.beta();

  particle1.k(0);
  particle2.k(0);
  
  double theta = acos(2*random.Uniform() - 1);
  double phi = TMath::Pi()*(2*random.Uniform() - 1);
  Fission(particle1, particle2, erel, theta, phi);

  LorentzBoost(particle1, -beta, 2);
  LorentzBoost(particle2, -beta, 2);
}


double CKinema::Scatter(double theta_lab, double phi_lab, CParticle a, CParticle b, CParticle& c, CParticle& d)
{
  double beta_lo = (a.p(2)+b.p(2))/(a.e()+b.e());

  LorentzBoost(a, beta_lo, 2);
  LorentzBoost(b, beta_lo, 2);

  Fission(c, d, a.e()+b.e());

  double theta_cm = ThetaLabToThetaCm(theta_lab, c.beta(), beta_lo);

  c.thetaphi(theta_cm, phi_lab);
  d.thetaphibar(theta_cm, phi_lab);
                
  LorentzBoost(c, -beta_lo, 2);
  LorentzBoost(d, -beta_lo, 2);

  return theta_cm;
}

double CKinema::ThetaLabToThetaCm(double thetalab, double betacm, double betalorentz)
{
  double gammalorentz = 1/sqrt(1-betalorentz*betalorentz);

  double p = gammalorentz * tan(thetalab);
  double q = betalorentz/betacm;

  double a = p*p + 1.0;
  double b = 2.0 * p*p * q;
  double c = p*p * q*q - 1.0;

  return acos((-b + sqrt(b*b - 4.0*a*c))/(2.0*a));
}

double CKinema::InvariantMass(const CParticle& particle1, const CParticle& particle2)
{
  double esum2 = (particle1.e()+particle2.e())*(particle1.e()+particle2.e());
  double psum2 = 0;
  for(int i = 0; i < 3; ++i){
    psum2 += (particle1.p(i)+particle2.p(i))*(particle1.p(i)+particle2.p(i));
  }

  return sqrt(esum2 - psum2);
}

double CKinema::BetheBloch::Function(std::string pname, std::string tname, double kin)
{
  static const double PI(3.14159265358279);
  static const double Na(6.022e+23);
  static const double me(5.1099890331e-1);
  static const double re(2.817e-13);

  double M, z, Z, A, I, C0, a, m, X1, X0, rho;
  BetheBloch::ReadInParamater(pname, tname, M, z, Z, A, I, C0, a, m, X1, X0, rho);

  /* 入射粒子のエネルギーに関連する量 */
  double energy(kin + M); /* 全エネルギー(MeV) */
  double beta_2(kin*(energy+M)/pow(energy,2.0)); /* ベータの二乗(MeV) */
  double gamma_2(1.0/(1.0-beta_2)); /* ガンマの二乗 */
  double eta_2(beta_2*gamma_2); /* イータの二乗 */ 

  if(eta_2 < 0.01){
    return 0;
  }

  //double coefficient = 0.1535; // [MeV*cm2/g]
  double coefficient = 2.0*PI*Na*pow(re,2.0)*me;

  /* Wmax */
  double s(me/M); /* 入射粒子と電子の質量比 */
  double Wmax((2.0*me*eta_2)/(1.0+2.0*s*sqrt(1.0+eta_2)+pow(s,2.0)));  /* 一回の衝突においてあたえるエネルギーの最大値 (MeV)*/

  /* 補正量deltaの計算(分極することにより外側の影響が減る効果) */
  double X = log10(beta_2*gamma_2)/2.0;
  double delta;
  if(X<X0){
    delta = 0.0;
  }else if(X>X0 && X<X1){
    delta = 4.6052*X+C0+a*pow(X1-X,m);
  }else if(X>X1){
    delta = 4.6052*X+C0;
  }else{
    std::cout << "paramater is wrong" << std::endl;
    abort();
  }
  
  /* 補正量C(I,eta)の計算(まわりの電荷が静止しているように見えなくなる効果) */
  double eta2 = 1.0/eta_2;
  double eta4 = eta2*eta2;
  double eta6 = eta2*eta4;
  double Ca = (0.422377*eta2+0.0304043*eta4-0.00038106*eta6)*pow(I,2.0)*1.0e+6;
  double Cb = (3.850190*eta2-0.1667989*eta4+0.00157955*eta6)*pow(I,3.0)*1.0e+9;
  double C = Ca+Cb;

  return coefficient*rho*Z/A*pow(z,2.0)/beta_2*(log(2.0*me*gamma_2*beta_2*Wmax/(I*I))-2.0*beta_2-delta-2.0*C/Z);
}

double CKinema::BetheBloch::Range(std::string pname, std::string tname,
				  double ein, double step)
{
  double del_e, BB;
  double energy = ein, range = 0;

  while(1){
    BB = Function(pname, tname, energy);
    if(BB == 0) break;
    del_e = energy/step;
    range += del_e/BB;
    energy -= del_e;
  }

  return range;
}

double CKinema::BetheBloch::EnergyOut(std::string pname, std::string tname,
				      double ein, double rstop, double step)
{
  double del_e = ein/step;
  double energy, range;
  for(energy = ein, range = 0; range < rstop; energy -= del_e){
    double BB = Function(pname, tname, energy);
    if(BB == 0){
      energy = 0;
      break;
    }else{
      del_e = energy/step;
      range += del_e/BB;
    }
  }

  return energy;
}

double CKinema::BetheBloch::TimeOut(std::string pname, std::string tname,
				    CParticle& particle, double rstop, double step)
{
  double energy=particle.k(), range=0, time=0;
  double del_e = energy/step;
  for(; range < rstop; energy -= del_e){
    double BB = Function(pname, tname, energy);
    if(BB == 0){
      time = -1.;
      break;
    }else{
      del_e = energy/step;
      range += del_e/BB;
      particle.k(energy);
      time += del_e/BB/particle.beta()/(TMath::C()*1e-7);
    }
  }

  return time;
}

void CKinema::BetheBloch::ReadInParamater(std::string pname, std::string tname,
		     double& M, double& z, double& Z, double& A,
		     double& I, double& C0, double& a, double& m,
		     double& X1, double& X0, double& rho)
{
  if(pname == "p"){
    M = 938.272; /* 入射粒子の質量(MeV) */
    z = 1.0; /* 入射粒子の電荷（e） */
  }else if(pname == "mu"){  
    M = 105.658;
    z = 1.0;
  }else{
    std::cout << "not yet defined " << pname << std::endl;
  }

  if(tname == "Scint"){
    //deltaが正確ではないが、近似としては比較的正しいはず
    double aH=1.104;
    double aC=1.;    
    double asum=aH+aC;
    Z = (aH*1+aC*6)/asum; /* 減衰させる物質の原子番号 */
    A = (aH*1+aC*12)/asum; /* 減衰させる物質の原子量 */
    I = 64.7e-6; /* mean excitation potential I(MeV) */
    C0 = -3.20;
    a = 0.1610;
    m = 3.24;
    X1 = 2.49;
    X0 = 0.1464;
    rho = 1.032; /* 密度(g/cm3) */
  }else if(tname == "Cu"){
    Z = 29.0;
    A = 63.546;
    I = 322.0e-6;
    C0 = -4.42;
    a = 0.1434;
    m = 2.9;
    X1 = 3.28;
    X0 = -0.0254;
    rho = 8.9;
  }else if(tname == "Pb"){
    Z = 82.0;
    A = 207.19;
    I = 823.0e-6;
    C0 = -6.2;
    a = 0.0936;
    m = 3.16;
    X1 = 3.81;
    X0 = 0.3776;
    rho = 11.34; 
  }else{
    std::cout << "not yet defined " << tname << std::endl;    
  }
}

double CKinema::MeVtoMeVee(double Tin, double Tout, std::string name)
{
  double a1, a2, a3, a4;
  double Tein, Teout, dTe;

  if(name == "e-" || name == "e+" || name == "mu-" || name == "mu+"){//leptons
    a1 = 1;
    a2 = 0;
    a3 = 0;
    a4 = 0;
  }else if(name == "proton" || name == "deuteron" || name == "triton"){//Z=1
    //NE-102(CECIL et al.)
    a1 = 0.95;
    a2 = 8.0;
    a3 = 0.1;
    a4 = 0.9;
  }else if(name == "alpha" || name == "He3"){//Z=2
    //plastic(CECIL et al.)
    a1 = 0.41;
    a2 = 5.9;
    a3 = 0.065;
    a4 = 1.01;
  }else{//heavy ions
    //used in DEMONS
    a1 = 0.29;
    a2 = 37.492;
    a3 = 0.01;
    a4 = 0.874;
  }

  Tein = a1*Tin-a2*(1-exp(-a3*pow(Tin, a4)));
  Teout = a1*Tout-a2*(1-exp(-a3*pow(Tout, a4)));
  dTe = Tein - Teout;

  if(dTe < 0) dTe = 0;

  return dTe;
}


