
#include <iostream>
#include <iomanip>
#include <cmath>

#include "TROOT.h"
#include "TMath.h"
#include "TVector3.h"
#include "TLorentzVector.h"

#include "CParticle.hh"

ClassImp(CParticle)

CParticle::CParticle(double mass, std::string name) : fName(name)
{
  if(mass < 0){
    std::cout << "mass must be non-negative" << std::endl;
    std::cout << "set mass as positive" << std::endl;
  }

  fP.SetVectM(TVector3(0., 0., 0.), std::abs(mass));

  fC=1;
  //  fC=29.9792458; // cm/ns
}

CParticle::~CParticle(){/* nothing */;}

void CParticle::p(double px, double py, double pz){SetMomentum(px, py, pz);}
void CParticle::pabs(double momentum){SetMomentum(momentum);}
void CParticle::e(double energy){SetEnergy(energy);}
void CParticle::k(double kineticenergy){SetKineticEnergy(kineticenergy);}
void CParticle::thetaphi(double ltheta, double lphi){SetThetaPhi(ltheta, lphi);}
void CParticle::thetaphibar(double ltheta, double lphi){SetThetaPhiBar(ltheta, lphi);}
void CParticle::beta(double bx, double by,double bz){SetBeta(bx, by, bz);}
void CParticle::betaabs(double lbeta){SetBeta(lbeta);}
void CParticle::c(double lightvelocity){SetC(lightvelocity);}

double CParticle::p(int i) const {return GetMomentum(i);}
double CParticle::p() const {return GetMomentum();}
double CParticle::p2() const {return p()*p();}
double CParticle::m() const {return GetMass();}
double CParticle::m2() const {return m()*m();}
double CParticle::e() const {return GetEnergy();}
double CParticle::e2() const {return e()*e();}
double CParticle::k() const {return GetKineticEnergy();}
double CParticle::k2() const {return k()*k();}
double CParticle::theta() const {return GetTheta();}
double CParticle::phi() const {return GetPhi();}
double CParticle::thetabar() const {return GetThetaBar();}
double CParticle::phibar() const {return GetPhiBar();}
double CParticle::beta(int i) const {return GetBeta(i);}
double CParticle::beta() const {return GetBeta();}
double CParticle::gamma(int i) const {return GetGamma(i);}
double CParticle::gamma() const {return GetGamma();}
double CParticle::v(int i) const {return GetVelocity(i);}
double CParticle::v() const {return GetVelocity();}
double CParticle::c() const {return GetC();}
double CParticle::c2() const {return c()*c();}

void CParticle::SetMomentum(double px, double py, double pz)
{
  fP.SetVectM(TVector3(px, py, pz), m());
}

void CParticle::SetMomentum(double momentum)
{
  if(momentum < 0){
    std::cout << "momentum must be non-negative" << std::endl;
    return;
  }

  double ltheta = theta();
  double lphi = phi();

  TVector3 p3;
  p3.SetX(momentum*sin(ltheta)*cos(lphi));
  p3.SetY(momentum*sin(ltheta)*sin(lphi));
  p3.SetZ(momentum*cos(ltheta));
  fP.SetVectM(p3, m());
}

void CParticle::SetEnergy(double energy)
{
  if(energy < m()*c2()){
    std::cout << "energy must larger than mass:" << m() << std::endl;
    return;
  }

  pabs(sqrt((energy-m()*c2())*(energy+m()*c2()))/c());
}

void CParticle::SetKineticEnergy(double kineticenergy)
{
  if(kineticenergy < 0){
    std::cout << "kinetic energy must be non-negative" << std::endl;
    return;
  }

  pabs(sqrt(2*m()*c2()*kineticenergy+kineticenergy*kineticenergy)/c());
}

void CParticle::SetThetaPhi(double ltheta, double lphi)
{
  if(ltheta < 0 || ltheta > TMath::Pi()){
    std::cout << "theta is out of range" << std::endl;
    return;
  }
  if(lphi <= -TMath::Pi() || lphi > TMath::Pi()){
    std::cout << "phi is out of range" << std::endl;
    return;
  }

  TVector3 p3;
  p3.SetX(p()*sin(ltheta)*cos(lphi));
  p3.SetY(p()*sin(ltheta)*sin(lphi));
  p3.SetZ(p()*cos(ltheta));
  fP.SetVectM(p3, m());
}

void CParticle::SetThetaPhiBar(double ltheta, double lphi)
{
  if(ltheta < 0 || ltheta > TMath::Pi()){
    std::cout << "theta is out of range" << std::endl;
    return;
  }
  if(lphi <= -TMath::Pi() || lphi > TMath::Pi()){
    std::cout << "phi is out of range" << std::endl;
    return;
  }

  ltheta = ThetaBar(ltheta);
  lphi = PhiBar(lphi);

  TVector3 p3;
  p3.SetX(p()*sin(ltheta)*cos(lphi));
  p3.SetY(p()*sin(ltheta)*sin(lphi));
  p3.SetZ(p()*cos(ltheta));
  fP.SetVectM(p3, m());
}

void CParticle::SetBeta(double bx, double by, double bz)
{
  double lbeta = sqrt(bx*bx+by*by+bz*bz);
  double px = m()*c2()*bx*sqrt(1/(1+lbeta)/(1-lbeta));
  double py = m()*c2()*by*sqrt(1/(1+lbeta)/(1-lbeta));
  double pz = m()*c2()*bz*sqrt(1/(1+lbeta)/(1-lbeta));

  p(px, py, pz);
}

void CParticle::SetBeta(double lbeta)
{
  if(lbeta < 0){
    std::cout << "beta must be non-negative" << std::endl;
    return;
  }

  pabs(m()*c2()*lbeta*sqrt(1/(1+lbeta)/(1-lbeta)));
}

void CParticle::SetC(double lightvelocity)
{
  double oldc = c();
  fC = lightvelocity;
  fP.SetVect(oldc/fC*fP.Vect());
}

double CParticle::GetMass() const
{
  return fP.M();
}

double CParticle::GetMomentum() const
{
  return fP.P();
}

double CParticle::GetMomentum(int i) const
{
  if(i < 0 || i > 3){
    std::cout << "only 0,1,2 for direction" << std::endl;
    return 0;
  }

  return fP(i);
}

double CParticle::GetEnergy() const
{
  return fP.E();
}

double CParticle::GetKineticEnergy() const
{
  return e()-m()*c2();
}
  
double CParticle::GetTheta() const
{
  double ltheta;
  if(p() == 0.0){
    ltheta = 0.0;
  }else{
    ltheta = acos(p(2)/p());
  }

  return ltheta;
}

double CParticle::GetPhi() const
{
  double lphi;
  if(p(0) > 0.0){
    lphi = atan(p(1)/p(0));
  }else if(p(0) < 0.0){
    lphi = PhiBar(atan(p(1)/p(0)));
  }else{
    if(p(1) > 0.0){
      lphi = TMath::Pi()*0.5;
    }else if(p(1) < 0.0){
      lphi = -TMath::Pi()*0.5;
    }else{
      lphi = 0.0;
    }
  }

  return lphi;
}

double CParticle::GetThetaBar() const
{
  return ThetaBar(GetTheta());
}

double CParticle::GetPhiBar() const
{
  return PhiBar(GetPhi());
}

double CParticle::GetBeta() const
{
  TVector3 beta3(beta(0), beta(1), beta(2));

  return beta3.Mag();
}

double CParticle::GetBeta(int i) const
{
  if(i < 0 || i > 3){
    std::cout << "only 0,1,2 for direction" << std::endl;
    return 0;
  }

  return p(i)*c()/e();
}

double CParticle::GetGamma() const
{
  return 1/sqrt((1-beta())*(1+beta()));
}

double CParticle::GetGamma(int i) const
{
  return sqrt(1/(1-beta(i))/(1+beta(i)));
}

double CParticle::GetVelocity() const
{
  return beta()*c();
}

double CParticle::GetVelocity(int i) const
{
  return beta(i)*c();
}

double CParticle::GetC() const
{
  return fC;
}

std::string CParticle::GetName() const
{
  return fName;
}

void CParticle::Print() const
{
  std::cout << "particle name:" << fName << std::endl
	    << "p(0):" << p(0) << " p(1):" << p(1) << " p(2):" << p(2) << " p:" << p() << std::endl
	    << "m:" << m() << std::endl
	    << "e:" << e() << std::endl    
	    << "k:" << k() << std::endl    
	    << "theta:" << theta() << std::endl    
	    << "phi:" << phi() << std::endl    
	    << "b:" << beta() << std::endl    
	    << "g:" << gamma() << std::endl    
	    << "v(0):" << v(0) << " v(1):" << v(1) << " v(2):" << v(2) << " v:" << v() << std::endl
	    << "c:" << c() << std::endl;
}

double CParticle::ThetaBar(double ltheta) const
{
  if(ltheta < 0 || ltheta > TMath::Pi()){
    std::cout << "theta is out of range" << std::endl;
    return 0;
  }

  return TMath::Pi() - ltheta;
}

double CParticle::PhiBar(double lphi) const
{
  if(lphi <= -TMath::Pi() || lphi > TMath::Pi()){
    std::cout << "theta is out of range" << std::endl;
    return 0;
  }

  return lphi <= 0 ? lphi+TMath::Pi() : lphi-TMath::Pi();
}


